/*
 * golombCoding.cpp
 *
 *  Created on: Jan 30, 2009
 *      Author: jonathanmasci
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "golombCoding.h"
#include <math.h>

GolombCoding::GolombCoding(unsigned char *input, unsigned int length,
                           unsigned int M) {
  _input = input;
  _length = length;
  _M = M;
  printf("M = %i\n", _M);
  // check if _M is power of two
  int logM = log2(_M);
  if (pow(2, logM) == _M)
    _isPowerOfTwo = true;
  else
    _isPowerOfTwo = false;
}

GolombCoding::~GolombCoding() {
}

void GolombCoding::encode() {
  unsigned char *tmpInt = new unsigned char[4];
  unsigned int num;
  unsigned int i, j, q, k; // loop indexes
  unsigned int v = 1; // the mask used in truncated binary coding
  unsigned int b; // numbers of bit to encode the remainder
  unsigned int new_remainder; // used if remainder >= pow(2, b) - _M

  for (i = 0; i < _length; i += 4) {
    // get the next integer from input
    memset(tmpInt, 0, 4);
    for (j = 0; j < 4 && j + i < _length; ++j) {
      tmpInt[j] = _input[i + j];
    }
    num = *(uint *)tmpInt;
    unsigned int quotient = num / _M;
    unsigned int remainder = num % _M;
    // code the quotient in unary coding
    for (q = 0; q < quotient; ++q)
      _bitVector.push_back(true);
    _bitVector.push_back(false);
    // code the remainder in truncated binary coding
    if (_isPowerOfTwo) { // coding remainder as binary format (Rice Coding)
      for (unsigned int k = 0 ; k < log2(_M); ++k) {
        _bitVector.push_back(v & num);
        v = v << 1;
      }
    } else {
      v = 1;
      b = ceil(log2(_M));
      if (remainder < (pow(2, b) - _M)) {
        // code r in plain binary using b-1 bits
        for (k = 0 ; k < b - 1; ++k) {
          _bitVector.push_back(v & remainder);
          v = v << 1;
        }
      } else {
        // code the number r + 2^b - M in plain binary using b bits
        new_remainder = remainder + pow(2, b) - _M;
        for (k = 0 ; k < b; ++k) {
          _bitVector.push_back(v & new_remainder);
          v = v << 1;
        }
      }
    }
  }
  delete[] tmpInt;
}

unsigned char *GolombCoding::decode(unsigned int *out_len) {
  unsigned int i = 0; // main _bitVector loop counter
  unsigned int j = 0; // counter to get the remainder
  unsigned int k = 0; // output stream position
  unsigned int quotient = 0;
  unsigned int remainder = 0;
  unsigned int *decoded = new unsigned int; // the decoded integer
  unsigned int cutoff = pow(2, ceil(log2(_M))) - _M;
  unsigned int output_length = 4;
  unsigned char *ret = new unsigned char[output_length];

  while (i < _bitVector.size()) {
    *decoded = 0;
    remainder = 0;
    quotient = 0;

    // getting the quotient by counting the ones until a zero occurs
    while (_bitVector[i] == true && i < _bitVector.size()) {
      ++quotient;
      ++i;
    }
    ++i; // for the 0
    // getting the remainder (b-1 bits)
    for (j = 0; j < ceil(log2(_M)) - 1 && i < _bitVector.size(); ++j, ++i)
      if (_bitVector[i] == true)
        remainder += 1 << j;
    if (remainder > cutoff) { // get the next bit because of the unused codeword
      if (_bitVector[i] == true)
        remainder += 1 << j;
      remainder -= cutoff;
      ++i;
    }
    *decoded = quotient * _M + remainder;
    // check output string size
    if (k + 4 > output_length) {
      output_length *= 2;
      ret = (unsigned char *)realloc(ret, output_length * sizeof(*ret));
    }
    // write the decoded integer to the output string
    unsigned char *tmp = (unsigned char *)decoded;
    ret[k] = tmp[0];
    ret[k + 1] = tmp[1];
    ret[k + 2] = tmp[2];
    ret[k + 3] = tmp[3];
    k += 4;
  }
  delete decoded;

  *out_len = k;
  return ret;
}

long GolombCoding::getBitSizeEstimation() {
  return _bitVector.size();
}
